import bs
import bsUtils
import bsSpaz
import bsFlag
import bsMap
import random
import bsPowerup
import bsScoreSet
import bsBomb
import weakref
import copy
import bsUI
import bsAchievement
import bsGame
import bsInternal
import bsTutorial

# team info
gTeamColors = ((0.2,0.4,1.6),)
gTeamNames = ("Good Guys",)
gMaxTeams = len(gTeamColors)

_gCampaigns = {}
def _registerCampaign(c):
    _gCampaigns[c._name] = c

def getCampaign(name): return _gCampaigns[name]

class Campaign(object):

    def __init__(self,name,sequential=True):
        self._name = name
        self._levels = []
        self._sequential = sequential

    def _getConfigDict(self):
        if not 'Campaigns' in bs.config:
            bs.config['Campaigns'] = {}
        if not self._name in bs.config['Campaigns']:
            bs.config['Campaigns'][self._name] = {}
        return bs.config['Campaigns'][self._name]

    def getName(self):
        return self._name

    def isSequential(self):
        return self._sequential

    def addLevel(self,name,gameType,config,scoreVersion=''):
        # scoreVersion is used if we want to reset score lists on an existing level
        # (if old scores would be out of range or whatnot)
        self._levels.append({'name':name,'gameType':gameType,'config':config,'scoreVersion':scoreVersion})

    def completeLevel(self,levelName):
        config = self._getConfigDict()
        if not levelName in config: config[levelName] = {'Rating':0.0,'Complete':False}
        state = config[levelName]
        state['Complete'] = True
        bs.writeConfig()
        
    def getLevels(self):
        levels = []
        for i,l in enumerate(self._levels):
            levels.append({'index':i,'name':l['name'],'Rating':None,'Complete':False})

        # now look at our prefs and fill in any info we have
        config = self._getConfigDict()
        levelDict = dict([[level['name'],level] for level in levels])
        for name,state in config.items():
            if name in levelDict:
                df = levelDict[name]
                try: df['Rating'] = state['Rating']
                except Exception: df['Rating'] = 0.0
                try: df['Complete'] = state['Complete']
                except Exception: df['Complete'] = False
            
        return levels

    def _getLevel(self,name):
        for l in self._levels:
            if l['name'] == name: return l
        raise Exception("level '"+name+"' not found.")

    def getLevelDisplayName(self,levelName):
        #name = self._getLevel(levelName)['displayName']
        name = levelName
        try: translated = bs.getResource('translations.coopLevelNames')[name]
        except Exception: translated = None
        if translated is None: translated = name
        return translated

    def getLevelGameType(self,levelName):
        return self._getLevel(levelName)['gameType']

    def getLevelSettings(self,levelName):
        settings = self._getLevel(levelName)['config']
        settings['name'] = levelName
        return settings
        
    def getLevelHighScores(self,levelName):
        level = self._getLevel(levelName)
        config = self._getConfigDict()
        highScoresKey = 'High Scores'+level['scoreVersion']
        if not levelName in config: return {}
        if not highScoresKey in config[levelName]: return {}
        return copy.deepcopy(config[levelName][highScoresKey])

    def getLevelScoreVersion(self,levelName):
        level = self._getLevel(levelName)
        return level['scoreVersion']

    def setLevelHighScores(self,levelName,levelHighScores):
        level = self._getLevel(levelName)
        config = self._getConfigDict()
        if not levelName in config: config[levelName] = {'Rating':0.0,'Complete':False}
        highScoresKey = 'High Scores'+level['scoreVersion']
        config[levelName][highScoresKey] = levelHighScores
        bs.writeConfig()

    def setSelectedLevel(self,levelName):
        self._getConfigDict()['Selection'] = levelName
        bs.writeConfig()

    def submitLevelRating(self,levelName,rating):
        config = self._getConfigDict()
        if not levelName in config: config[levelName] = {'Rating':0.0,'Complete':False}
        state = config[levelName]
        state['Rating'] = max(state['Rating'],rating)
        bs.writeConfig()

    def getLevelRating(self,levelName):
        config = self._getConfigDict()
        if not levelName in config: return None
        state = config[levelName]
        return state['Rating']
            
    def getSelectedLevel(self):
        try: return self._getConfigDict()['Selection']
        except Exception: return self._levels[0]['name']

    def reset(self):
        self._getConfigDict() # make sure path exists..
        bs.config['Campaigns'][self._name] = {}
        bs.writeConfig()


class CoopScoreScreen(bs.Activity):

    def __init__(self,session,settings={}):
        bs.Activity.__init__(self,session,settings=settings)
        self._transitionTime = 500 # keeps prev activity alive while we fade in

        self._goldTex = bs.getTexture('medalGold')
        self._silverTex = bs.getTexture('medalSilver')
        self._bronzeTex = bs.getTexture('medalBronze')

        self.scoreDisplaySound = bs.getSound("scoreHit01")
        self.scoreDisplaySoundSmall = bs.getSound("scoreHit02")
        self.drumRollSound = bs.getSound('drumRoll')
        self.cymbalSound = bs.getSound('cymbal')
        self._menuIconTexture = bs.getTexture('menuIcon')
        gs = bsInternal._getGameService()
        if gs == 'Game Center': self._gameServiceTexture = bs.getTexture('gameCenterIcon')
        elif gs == 'Game Circle': self._gameServiceTexture = bs.getTexture('gameCircleIcon')
        else: self._gameServiceTexture = None
        self._replayIconTexture = bs.getTexture('replayIcon')
        self._nextLevelIconTexture = bs.getTexture('nextLevelIcon')
        self._cashRegisterSound = bs.getSound('cashRegister')
        self._gunCockingSound = bs.getSound('gunCocking')
        self._dingSound = bs.getSound('ding')
        self._scoreLink = None
        self._rootUI = None

    def __del__(self):
        bs.Activity.__del__(self)
        # if our UI is still up, kill it
        if self._rootUI is not None and self._rootUI.exists():
            bs.containerWidget(edit=self._rootUI,transition='outLeft')

    def onTransitionIn(self):
        bs.Activity.onTransitionIn(self)
        self._background = bsUtils.Background(self,bs.Actor.UNRETAINED,
                                              fadeTime=450,startFaded=False,showLogo=True)

    def _uiMenu(self):
        bs.containerWidget(edit=self._rootUI,transition='outLeft')
        bs.gameTimer(100,bs.WeakCall(self.getSession().end))

    def _uiRestart(self):
        if bsInternal._getDemoTimeRemaining() == 0:
            bsUI.showDemoExpiredUI()
            return
        # any time we're restarting, we set this level as current
        self._campaign.setSelectedLevel(self._level)
        bs.containerWidget(edit=self._rootUI,transition='outLeft')
        self.end({'outcome':'restart'})

    def _uiNext(self):
        if bsInternal._getDemoTimeRemaining() == 0:
            bsUI.showDemoExpiredUI()
            return

        # if we didn't just complete this level but are choosing
        # to play the next one, set it as current
        # (this won't happen otherwise)
        if self._isComplete and self._isMoreLevels and not self._newlyComplete:
            self._campaign.setSelectedLevel(self._nextLevelName)
        bs.containerWidget(edit=self._rootUI,transition='outLeft')
        self.end({'outcome':'nextLevel'})

    def _uiGC(self):
        bsInternal._showOnlineScoreUI('leaderboard',game=self._gameNameStr,gameVersion=self._gameConfigStr)

    def _uiShowAchievements(self):
        bsInternal._showOnlineScoreUI('achievements')

    def _uiWorldsBest(self):
        if self._scoreLink == None:
            bs.playSound(bs.getSound('error'))
            bs.screenMessage(bs.getResource('scoreListUnavailableText'),color=(1,0.5,0))
        else:
            bs.openURL(self._scoreLink)

    def _uiError(self):
        self._nextLevelError = bsUtils.Text(self,bs.Actor.UNRETAINED,
                                            bs.getResource('completeThisLevelToProceedText'),
                                            flash=True,
                                            maxWidth=360,
                                            scale=0.54,
                                            hAlign='center',
                                            color=(0.5,0.7,0.5,1),
                                            position=(300,-235))
        bs.playSound(bs.getSound('error'))
        bs.gameTimer(2000,bs.WeakCall(self._nextLevelError.handleMessage,bs.DieMessage()))

    def onPlayerLeaving(self,player):
        bs.Activity.onPlayerLeaving(self,player)
        # if all our players leave we wanna quit out of the game
        bs.realTimer(2000,bs.WeakCall(self._endSessionIfEmpty))
        
    def _endSessionIfEmpty(self):
        if len(self.players) == 0:
            self.getSession().end()

    def requestUI(self):
        # we don't want to just show our UI in case the user already has the main menu up
        # so instead we add a callback for when the menu goes away; if we're still alive, we'll come up then
        # if there's no menu this gets called immediately
        bsUI.addMainMenuCloseCallback(bs.WeakCall(self.showUI))

    def showUI(self):
        delay = 700 if (self._score is not None) else 0

        # if there's no players left in the game, lets not show the UI
        # (that would allow restarting the game with zero players, etc)

        # hmm shouldnt need this try/except here i dont think..
        try: players = self.players
        except Exception,e:
            print 'EXC bsCoopGame showUI cant get self.players; shouldnt happen:',e
            players = []
        if len(players) == 0: return

        d = self._rootUI = bs.containerWidget(size=(0,0),transition='inLeft')

        hOffs = 7
        vOffs = -280

        # we wanna prevent controllers users from popping up browsers or game-center widgets in cases
        # where they cant easily get back to the game (like on mac)
        canSelectExtraButtons = 'android' in bs.getEnvironment()['userAgentString']

        bsInternal._setUIInputDevice(None) # menu is up for grabs

        if self._haveFriendScores:
            gcButton = bs.buttonWidget(parent=d,color=(0.45,0.4,0.5),position=(hOffs-520,vOffs+480),size=(300,60),
                                       label=bs.getResource('topFriendsText'),call=bs.WeakCall(self._uiGC),transitionDelay=delay+500,
                                       icon=self._gameServiceTexture,
                                       selectable=canSelectExtraButtons)

        if bsInternal._getGameService() is not None:
            gcButton = bs.buttonWidget(parent=d,color=(0.45,0.4,0.5),position=(hOffs-520,vOffs+450-235+40),size=(300,60),
                                       label=bs.getResource('achievementsText'),call=bs.WeakCall(self._uiShowAchievements),transitionDelay=delay+1500,
                                       icon=self._gameServiceTexture,
                                       selectable=canSelectExtraButtons)

        if not bsInternal._isRunningOnFantasia():
            gcButton = bs.buttonWidget(parent=d,color=(0.45,0.4,0.5),position=(160,vOffs+480),size=(350,62),
                                       label=bs.getResource('worldsBestScoresText') if self._scoreType == 'points' else bs.getResource('worldsBestTimesText'),
                                       call=bs.WeakCall(self._uiWorldsBest),transitionDelay=delay+1900,selectable=canSelectExtraButtons)
        else:
            pass

        if not self._isMoreLevels:
            hOffs += 70

        menuButton = bs.buttonWidget(parent=d,position=(hOffs-130-60,vOffs),size=(110,85),label='',call=bs.WeakCall(self._uiMenu))
        i = bs.imageWidget(parent=d,drawController=menuButton,position=(hOffs-130-60+22,vOffs+14),size=(60,60),texture=self._menuIconTexture,opacity=0.8)
        restartButton = bs.buttonWidget(parent=d,position=(hOffs-60,vOffs),size=(110,85),label='',call=bs.WeakCall(self._uiRestart))
        i = bs.imageWidget(parent=d,drawController=restartButton,position=(hOffs-60+19,vOffs+7),size=(70,70),texture=self._replayIconTexture,opacity=0.8)

        # our 'next' button is disabled if we havn't unlocked the next level yet and invisible if there is none
        if self._isMoreLevels:
            if self._isComplete:
                call = bs.WeakCall(self._uiNext)
                buttonSound = True
                imageOpacity = 0.8
                color = None
            else:
                call = bs.WeakCall(self._uiError)
                buttonSound = False
                imageOpacity = 0.2
                color = (0.3,0.3,0.3)
            nextButton = bs.buttonWidget(parent=d,position=(hOffs+130-60,vOffs),size=(110,85),label='',call=call,color=color,enableSound=buttonSound)
            i = bs.imageWidget(parent=d,drawController=nextButton,position=(hOffs+130-60+12,vOffs+5),size=(80,80),texture=self._nextLevelIconTexture,opacity=imageOpacity)

        if self._newlyComplete and self._victory and self._isMoreLevels:
            bs.containerWidget(edit=d,selectedChild=nextButton)

        else: bs.containerWidget(edit=d,selectedChild=restartButton)
        
    def onBegin(self):

        bs.Activity.onBegin(self)

        self._playerInfo = self.settings['playerInfo']
        self._score = self.settings['score']
        self._failMessage = self.settings['failMessage']

        self._beginTime = bs.getGameTime()

        if 'scoreOrder' in self.settings:
            if not self.settings['scoreOrder'] in ['increasing','decreasing']:
                raise Exception("Invalid score order: "+self.settings['scoreOrder'])
            self._scoreOrder = self.settings['scoreOrder']
        else: self._scoreOrder = 'increasing'

        if 'scoreType' in self.settings:
            if not self.settings['scoreType'] in ['points','time']:
                raise Exception("Invalid score type: "+self.settings['scoreType'])
            self._scoreType = self.settings['scoreType']
        else: self._scoreType = 'points'

        self._campaign = self.settings['campaign']
        self._level = self.settings['level']

        self._gameNameStr = self._campaign.getName()+":"+self._level
        self._gameConfigStr = str(len(self._playerInfo))+"p"+self._campaign.getLevelScoreVersion(self._level).replace(' ','_')

        # if game-center/etc scores are available we show our friends' scores
        # otherwise we show our local high scores
        self._haveFriendScores = bsInternal._gameServiceHasLeaderboard(self._gameNameStr,self._gameConfigStr)

        try: self._oldBestRank = self._campaign.getLevelRating(self._level)
        except Exception: self._oldBestRank = 0.0

        # calc whether the level is complete and other stuff
        victory = (self.settings['outcome'] == 'victory')
        levels = self._campaign.getLevels()
        level = None
        for l in levels:
            if l['name'] == self._level:
                level = l
                break
        if level == None: raise Exception()
        self._victory = victory
        self._wasComplete = level['Complete']
        self._isComplete = (self._wasComplete or victory)
        self._newlyComplete = (self._isComplete and not self._wasComplete)
        self._isMoreLevels = (level['index'] < len(levels)-1) and self._campaign.isSequential()

        # any time we complete a level, set the next one as unlocked
        if self._isComplete and self._isMoreLevels:
            self._campaign.completeLevel(self.settings['level'])
            self._nextLevelName = levels[level['index']+1]['name']

            # if this is the first time we completed it, set the next one as current
            if self._newlyComplete:
                bs.config['Selected Coop Game'] = self._campaign.getName()+":"+self._nextLevelName
                bs.writeConfig()
                self._campaign.setSelectedLevel(self._nextLevelName)

        bs.gameTimer(1000,bs.WeakCall(self.requestUI))

        if self._isComplete and self._victory and self._isMoreLevels:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('levelUnlockedText')+":\n" if self._newlyComplete else bs.getResource('nextLevelText')+":\n",
                         transition='inRight',
                         transitionDelay=5200,
                         flash=self._newlyComplete,
                         scale=0.54,
                         hAlign='center',
                         maxWidth=270,
                         color=(0.5,0.7,0.5,1),
                         position=(270,-235))
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.translate('coopLevelNames',self._nextLevelName),
                         transition='inRight',
                         transitionDelay=5200,
                         flash=self._newlyComplete,
                         scale=0.7,
                         hAlign='center',
                         maxWidth=205,
                         color=(0.5,0.7,0.5,1),
                         position=(270,-255))
            if self._newlyComplete:
                bs.gameTimer(5200,bs.Call(bs.playSound,self._cashRegisterSound))
                bs.gameTimer(5200,bs.Call(bs.playSound,self._dingSound))

        offsX = -195
        #if len(self._playerInfo) > 1: pstr = '- '+str(len(self._playerInfo))+' players -'
        if len(self._playerInfo) > 1: pstr = '- '+bs.getResource('multiPlayerCountText').replace('${COUNT}',str(len(self._playerInfo)))+' -'
        else: pstr = '- '+bs.getResource('singlePlayerCountText')+' -'
        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                         bs.translate('coopLevelNames',self._level),
                         maxWidth=800,
                         flash=False,trail=False,
                         color=(0.5,1,0.5,1),
                         hAlign='center',
                         scale=0.4,
                         position=(0,292),
                         jitter=1.0)
        bsUtils.Text(self,bs.Actor.RETAINED,
                     pstr,
                     maxWidth=300,
                     transition='fadeIn',
                     scale=0.7,
                     hAlign='center',
                     vAlign='center',
                     color=(0.5,0.7,0.5,1),
                     position=(0,230))
        if self._score != None:
            bs.gameTimer(350,bs.Call(bs.playSound,self.scoreDisplaySoundSmall))


        # vestigial remain.. this stuff should just be instance vars
        self._showInfo = {}

        if self._score != None:
            bs.gameTimer(800,bs.WeakCall(self._showScoreVal,offsX))
        else:
            bs.gameTimer(1,bs.WeakCall(self._showFail))


        self._nameStr = nameStr = ', '.join([p['name'] for p in self._playerInfo])

        if self._haveFriendScores:
            self._friendsLoadingStatus = bsUtils.Text(self,bs.Actor.UNRETAINED,
                                                      bs.getResource('loadingText')+'...',
                                                      #'loading...',
                                                      position=(-405,150+30),
                                                      color=(1,1,1,0.4),
                                                      transition='fadeIn',
                                                      scale=0.7,
                                                      transitionDelay=2000)
        self._scoreLoadingStatus = bsUtils.Text(self,bs.Actor.UNRETAINED,
                                                bs.getResource('loadingText')+'...',
                                                #'loading...',
                                                position=(280,150+30),
                                                color=(1,1,1,0.4),
                                                transition='fadeIn',
                                                scale=0.7,
                                                transitionDelay=2000)

        bsInternal._submitScore(self._gameNameStr,self._gameConfigStr,nameStr,-1 if self._score == None else self._score,
                       bs.WeakCall(self._gotScoreResults),
                       bs.WeakCall(self._gotFriendScoreResults) if self._haveFriendScores else None,
                       order=self._scoreOrder)

        if self._score != None:
            bs.gameTimer(400,bs.WeakCall(self._playDrumRoll))

        # add us to high scores, filter, and store
        ourHighScoresAll = self._campaign.getLevelHighScores(self._level)
        try: ourHighScores = ourHighScoresAll[str(len(self._playerInfo))+" Player"]
        except Exception: ourHighScores = ourHighScoresAll[str(len(self._playerInfo))+" Player"] = []

        if self._score != None:
            ourScore = (self._score,{'players':self._playerInfo})
            ourHighScores.append(ourScore)
            ourHighScores.sort(reverse=True if self._scoreOrder == 'increasing' else False)
        else: ourScore = None

        del ourHighScores[10:]

        if self._score != None:
            self._campaign.setLevelHighScores(self._level,ourHighScoresAll)


        # if we're not doing the world's-best button
        if bsInternal._isRunningOnFantasia():
            tsHeight = 300
            tsHOffs = 210
            vOffs = 40
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('worldsBestScoresText') if self._scoreType == 'points' else bs.getResource('worldsBestTimesText'),
                         maxWidth=210,
                         position=(tsHOffs-10,tsHeight/2+25+vOffs+20),
                         transition='inLeft',
                         vAlign='center',
                         scale=1.2,
                         transitionDelay=2200)
            
        # display these if we're not gonna try to do friend scores
        if not self._haveFriendScores:

            tsHeight = 300
            tsHOffs = -480
            vOffs = 40
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('yourBestScoresText') if self._scoreType == 'points' else bs.getResource('yourBestTimesText'),
                         maxWidth=210,
                         position=(tsHOffs-10,tsHeight/2+25+vOffs+20),
                         transition='inRight',
                         vAlign='center',
                         scale=1.2,
                         transitionDelay=1800)

            displayScores = list(ourHighScores)

            displayCount = 5

            while len(displayScores) < displayCount: displayScores.append((0,None))

            showedOurs = False

            hOffsExtra = 85 if self._scoreType == 'points' else 130
            vOffsExtra = 20
            vOffsNames = 0
            scale = 1.0
            pCount = len(self._playerInfo)

            hOffsExtra -= 75
            if pCount > 1:
                hOffsExtra -= 20
            if pCount == 2:
                #vOffsNames = 2
                scale = 0.9
            elif pCount == 3:
                #vOffsNames = 5
                scale = 0.65
            elif pCount == 4:
                #vOffsNames = 7
                scale = 0.5

            times = []

            for i in range(displayCount):
                times.insert(random.randrange(0,len(times)+1),(1900+i*50,2300+i*50))

            for i in range(displayCount):
                score = displayScores[i][0]
                try: nameStr = ', '.join([p['name'] for p in displayScores[i][1]['players']])
                except Exception:
                    nameStr = '-'
                if displayScores[i] == ourScore and not showedOurs:
                    flash = True
                    color0 = (0.6,0.4,0.1,1.0)
                    color1 = (0.6,0.6,0.6,1.0)
                    tDelay1 = 3700
                    tDelay2 = 3700
                    showedOurs = True
                else:
                    flash = False
                    color0 = (0.6,0.4,0.1,1.0)
                    color1 = (0.6,0.6,0.6,1.0)
                    tDelay1 = times[i][0]
                    tDelay2 = times[i][1]
                bsUtils.Text(self,bs.Actor.RETAINED,
                             str(displayScores[i][0]) if self._scoreType == 'points' else bsUtils.getTimeString(displayScores[i][0]*10),
                             position=(tsHOffs+20+hOffsExtra,vOffsExtra+tsHeight/2+-tsHeight*(i+1)/10+vOffs+11.0),
                             hAlign='right',
                             vAlign='center',
                             color=color0,
                             flash=flash,
                             transition='inRight',
                             transitionDelay=tDelay1)

                bsUtils.Text(self,bs.Actor.RETAINED,
                             nameStr,
                             position=(tsHOffs+35+hOffsExtra,vOffsExtra+tsHeight/2+-tsHeight*(i+1)/10+vOffsNames+vOffs+11.0),
                             maxWidth = 80.0+100.0*len(self._playerInfo),
                             vAlign='center',
                             color=color1,
                             flash=flash,
                             scale=scale,
                             transition='inRight',
                             transitionDelay=tDelay2)

        # show achievements for this level
        tsHeight = -150
        tsHOffs = -480
        vOffs = 40
        if bsInternal._getGameService() is None:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('achievementsText'),
                         position=(tsHOffs-10,tsHeight/2+25+vOffs+3),
                         maxWidth=210,
                         transition='inRight',
                         vAlign='center',
                         scale=1.2,
                         transitionDelay=2800)
        achievements = bsAchievement.getAchievementsForCoopLevel(self._gameNameStr)
        h = -455
        v = -100
        tDelay = 0
        for a in achievements:
            a.createDisplay(self,h,v+vOffs,3000+tDelay)
            v -= 55
            tDelay += 250


        bs.gameTimer(5000,bs.WeakCall(self._showTips))

    def _playDrumRoll(self):
        bs.NodeActor(self,bs.Actor.RETAINED,
                     bs.newNode('sound',
                                self,
                                attrs={'sound':self.drumRollSound,
                                       'positional':False,
                                       'loop':False}))

    def _gotFriendScoreResults(self,results):

        # delay a bit if results come in too fast
        baseDelay = max(0,1900-(bs.getGameTime()-self._beginTime))

        tsHeight = 300
        tsHOffs = -550
        vOffs = 30

        # report in case of error
        if results is None:
            self._friendsLoadingStatus = bsUtils.Text(self,bs.Actor.UNRETAINED,
                                                      bs.getResource('friendScoresUnavailableText'),
                                                      maxWidth=330,
                                                      position=(-475,150+vOffs),
                                                      color=(1,1,1,0.4),
                                                      transition='fadeIn',
                                                      transitionDelay=baseDelay+800,
                                                      scale=0.7)
            return

        self._friendsLoadingStatus = None


        # ok, it looks like we aren't able to reliably get a just-submitted result
        # returned in the score list, so we need to look for our score in this list
        # and replace it if ours is better or add ours otherwise.
        if self._score != None:
            ourScoreEntry = [self._score,'Me',True]
            for score in results:
                if score[2]:
                    if self._scoreOrder == 'increasing': ourScoreEntry[0] = max(score[0],self._score)
                    else: ourScoreEntry[0] = min(score[0],self._score)
                    results.remove(score)
                    break
            results.append(ourScoreEntry)
            results.sort(reverse=True if self._scoreOrder == 'increasing' else False)
        # if we're not submitting our own score, we still want to change the name of our own score to 'Me'
        else:
            for score in results:
                if score[2]:
                    score[1] = 'Me'
                    break
            
        hOffsExtra = 80 if self._scoreType == 'points' else 130
        vOffsExtra = 20
        vOffsNames = 0
        scale = 1.0

        # make sure there's at least 5..
        while len(results) < 5:
            results.append([0,'-',False])
        results = results[:5]

        times = []

        for i in range(len(results)):
            times.insert(random.randrange(0,len(times)+1),(baseDelay+i*50,baseDelay+300+i*50))

        for i,t in enumerate(results):
            score = int(t[0])
            nameStr = t[1]
            isMe = t[2]
            # if self._showInfo['results']['rank'] == i+1:
            if isMe and score == self._score:
                flash = True
                color0 = (0.6,0.4,0.1,1.0)
                color1 = (0.6,0.6,0.6,1.0)
                tDelay1 = baseDelay+1000
                tDelay2 = baseDelay+1000
            else:
                flash = False
                if isMe:
                    color0 = (0.6,0.4,0.1,1.0)
                    color1 = (0.9,1.0,0.9,1.0)
                else:
                    color0 = (0.6,0.4,0.1,1.0)
                    color1 = (0.6,0.6,0.6,1.0)
                tDelay1 = times[i][0]
                tDelay2 = times[i][1]
            if nameStr != '-':
                bsUtils.Text(self,bs.Actor.RETAINED,
                             str(score) if self._scoreType == 'points' else bsUtils.getTimeString(score*10),
                             position=(tsHOffs+20+hOffsExtra,vOffsExtra+tsHeight/2+-tsHeight*(i+1)/10+vOffs+11.0),
                             hAlign='right',
                             vAlign='center',
                             color=color0,
                             flash=flash,
                             transition='inRight',
                             transitionDelay=tDelay1)
            else:
                if isMe: print 'Error: got empty nameStr on score result:',t
            bsUtils.Text(self,bs.Actor.RETAINED,
                         nameStr,position=(tsHOffs+35+hOffsExtra,vOffsExtra+tsHeight/2+-tsHeight*(i+1)/10+vOffsNames+vOffs+11.0),
                         color=color1,
                         maxWidth=160.0,
                         vAlign='center',
                         flash=flash,
                         scale=scale,
                         transition='inRight',
                         transitionDelay=tDelay2)

    def _gotScoreResults(self,results):
    
        # delay a bit if results come in too fast
        baseDelay = max(0,2700-(bs.getGameTime()-self._beginTime))
        vOffs = 20

        if results is None:
            self._scoreLoadingStatus = bsUtils.Text(self,bs.Actor.UNRETAINED,
                                                    bs.getResource('worldScoresUnavailableText'),
                                                    position=(210,150+vOffs+10),
                                                    color=(1,1,1,0.4),
                                                    transition='fadeIn',
                                                    transitionDelay=baseDelay+300,
                                                    scale=0.7)
        else:
            self._scoreLink = results['link']
            if not self._scoreLink.startswith('http://'):
                self._scoreLink = bsInternal._getServerAddress()+"/"+self._scoreLink
            self._scoreLoadingStatus = None

        self._showInfo['results'] = results
        if results != None:
            
            if results['tops'] != '':
                self._showInfo['tops'] = results['tops']

            else: self._showInfo['tops'] = []

            # we keep track of our best ranks per game
            if self._score != None:
                rank = self._showInfo['results']['rank']
                total = self._showInfo['results']['total']
                rating = 10.0 if total == 1 else round(10.0 * (1.0 - (float(rank-1)/(total-1))),1)
                self._campaign.submitLevelRating(self._level,rating)

        offsX = -195
        available = (self._showInfo['results'] != None)

        if self._score != None:
            bs.gameTimer(1500+baseDelay,bs.WeakCall(self._showWorldRank,offsX))

        tsHOffs = 200
        tsHeight = 300

        # show world tops
        if available:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('lastGamesText').replace('${COUNT}',str(self._showInfo['results']['total'])),
                         #"(last "+str(self._showInfo['results']['total'])+" games)",
                         position=(tsHOffs-35+95,tsHeight/2+6+vOffs),
                         color=(0.4,0.4,0.4,1.0),
                         scale=0.7,
                         transition='inRight',
                         transitionDelay=baseDelay+300)

            hOffsExtra = 0
            vOffsNames = 0
            scale = 1.0
            pCount = len(self._playerInfo)
            if pCount > 1:
                hOffsExtra -= 40
            if self._scoreType != 'points':
                hOffsExtra += 60

            if pCount == 2:
                #vOffsNames = 2
                scale = 0.9
            elif pCount == 3:
                #vOffsNames = 5
                scale = 0.65
            elif pCount == 4:
                #vOffsNames = 7
                scale = 0.5

            # make sure there's at least 10..
            while len(self._showInfo['tops']) < 10:
                self._showInfo['tops'].append([0,'-'])

            times = []

            for i in range(len(self._showInfo['tops'])):
                times.insert(random.randrange(0,len(times)+1),(baseDelay+i*50,baseDelay+400+i*50))

            for i,t in enumerate(self._showInfo['tops']):
                score = int(t[0])
                nameStr = t[1]

                if self._nameStr == nameStr and self._score == score:
                    flash = True
                    color0 = (0.6,0.4,0.1,1.0)
                    color1 = (0.6,0.6,0.6,1.0)
                    tDelay1 = baseDelay+1000
                    tDelay2 = baseDelay+1000
                else:
                    flash = False
                    if self._nameStr == nameStr:
                        color0 = (0.6,0.4,0.1,1.0)
                        color1 = (0.9,1.0,0.9,1.0)
                    else:
                        color0 = (0.6,0.4,0.1,1.0)
                        color1 = (0.6,0.6,0.6,1.0)
                    tDelay1 = times[i][0]
                    tDelay2 = times[i][1]
                if nameStr != '-':
                    bsUtils.Text(self,bs.Actor.RETAINED,
                                 str(score) if self._scoreType == 'points' else bsUtils.getTimeString(score*10),
                                 position=(tsHOffs+20+hOffsExtra,tsHeight/2+-tsHeight*(i+1)/10+vOffs+11.0),
                                 hAlign='right',
                                 vAlign='center',
                                 color=color0,
                                 flash=flash,
                                 transition='inLeft',
                                 transitionDelay=tDelay1)
                bsUtils.Text(self,bs.Actor.RETAINED,
                             nameStr,position=(tsHOffs+35+hOffsExtra,tsHeight/2+-tsHeight*(i+1)/10+vOffsNames+vOffs+11.0),
                             maxWidth=80.0+100.0*len(self._playerInfo),
                             vAlign='center',
                             color=color1,
                             flash=flash,
                             scale=scale,
                             transition='inLeft',
                             transitionDelay=tDelay2)

    def _showTips(self):
        bsUtils.TipsText(self,bs.Actor.RETAINED,offsY=30)

    def _showWorldRank(self,offsX):
        available = (self._showInfo['results'] != None)
        if available:
            rank = self._showInfo['results']['rank']
            total = self._showInfo['results']['total']
            rating = 10.0 if total == 1 else round(10.0 * (1.0 - (float(rank-1)/(total-1))),1)
        else: rating = None


        if available:
            if rating >= 9.0: medalTex = self._goldTex
            elif rating >= 6.0: medalTex = self._silverTex
            else: medalTex = self._bronzeTex
        else: medalTex = None

        medalTex = None
        if medalTex == None:
            medalOffset = 0
            medalAlign='center'
        else:
            medalOffset = -50
            medalAlign = 'left'
            bsUtils.Image(self,bs.Actor.RETAINED,
                          medalTex,position=(160+medalOffset+offsX,-147),
                          transition='fadeIn',
                          transitionDelay=1000,
                          color=(1,1,1,1.0),
                          scale=(60,60))

        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                         (str(rating) if available else bs.getResource('unavailableText')),
                         flash=True,trail=True,
                         scale = 0.6 if available else 0.333,
                         tiltTranslate=0.11,
                         hAlign='center',
                         position=(190+offsX,-94),
                         maxWidth=200,
                         jitter=1.0)

        if available:
            if rating >= 9.5: stars = 3
            elif rating >= 7.5: stars = 2
            else: stars = 1

            starTex = bs.getTexture('star')
            starX = 135+offsX
            for i in range(stars):
                img = bs.NodeActor(self,bs.Actor.RETAINED,
                                   bs.newNode('image',
                                              owner=self,
                                              attrs={'texture':starTex,
                                                     'translate':(starX,-16),
                                                     'scale':(62,62),
                                                     'opacity':1.0,
                                                     'color':(2.2,1.2,0.3),
                                                     'absoluteScale':True}))
                bsUtils.animate(img.node,'opacity',{150:0,400:1})
                starX += 60
            for i in range(3-stars):
                img = bs.NodeActor(self,bs.Actor.RETAINED,
                                   bs.newNode('image',
                                              owner=self,
                                              attrs={'texture':starTex,
                                                     'translate':(starX,-16),
                                                     'scale':(62,62),
                                                     'opacity':1.0,
                                                     'color':(0.3,0.3,0.3),
                                                     'absoluteScale':True}))
                bsUtils.animate(img.node,'opacity',{150:0,400:1})
                starX += 60
            
            def foo(count,x,offsY,score):
                bsUtils.Text(self,bs.Actor.RETAINED,
                             score+' =',
                             position=(x,-64+offsY),
                             color=(0.6,0.6,0.6,0.6),
                             hAlign='center',vAlign='center',transition='fadeIn',scale=0.4,
                             transitionDelay=1000)
                stx = x + 20
                for i in range(count):
                    img = bs.NodeActor(self,bs.Actor.RETAINED,
                                       bs.newNode('image',
                                                  owner=self,
                                                  attrs={'texture':starTex,
                                                         'translate':(stx,-64+offsY),
                                                         'scale':(12,12),
                                                         'opacity':0.7,
                                                         'color':(2.2,1.2,0.3),
                                                         'absoluteScale':True}))
                    bsUtils.animate(img.node,'opacity',{1000:0,1500:0.5})
                    stx += 13.0

            foo(1,-44-30,-112,'0.0')
            foo(2,10-30,-112,'7.5')
            foo(3,77-30,-112,'9.5')

            

        try: bestRank = self._campaign.getLevelRating(self._level)
        except Exception: bestRank = 0.0

        if available:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('outOfText').replace('${RANK}',str(int(self._showInfo['results']['rank']))).replace('${ALL}',str(self._showInfo['results']['total'])),
                         position=(medalOffset,-155 if self._newlyComplete else -145),
                         color=(1,1,1,0.7),
                         hAlign=medalAlign,
                         transition='fadeIn',
                         scale=0.55,
                         transitionDelay=1000)

        newBest = (bestRank > self._oldBestRank and bestRank > 0.0)
        wasString = '' if self._oldBestRank == None else ' '+bs.getResource('scoreWasText').replace('${COUNT}',str(self._oldBestRank))
        if not self._newlyComplete:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         (bs.getResource('newPersonalBestText')+wasString) if newBest else bs.getResource('bestRatingText').replace('${RATING}',str(bestRank)),
                         position=(medalOffset,-165),
                         color=(1,1,1,0.7),
                         flash=newBest,
                         hAlign=medalAlign,
                         transition='inRight' if newBest else 'fadeIn',
                         scale=0.5,
                         transitionDelay=1000)
        bsUtils.Text(self,bs.Actor.RETAINED,
                     bs.getResource('ratingText')+":",
                     position=(0,36),
                     maxWidth=300,
                     transition='fadeIn',
                     hAlign='center',
                     vAlign='center',
                     transitionDelay=0)
        bs.gameTimer(350,bs.Call(bs.playSound,self.scoreDisplaySound))
        bs.gameTimer(350,bs.Call(bs.playSound,self.cymbalSound))

    def _showFail(self):
        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                         bs.getResource('failText'),
                         maxWidth=300,
                         flash=False,trail=True,
                         hAlign='center',
                         tiltTranslate=0.11,
                         position=(0,40),
                         jitter=1.0)
        if self._failMessage != None:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         self._failMessage,
                         hAlign='center',
                         position=(0,-130),
                         maxWidth=300,
                         color=(1,1,1,0.5),
                         transition='fadeIn',
                         transitionDelay=1000)
        bs.gameTimer(350,bs.Call(bs.playSound,self.scoreDisplaySound))
        
    def _showScoreVal(self,offsX):
        bsUtils.ZoomText(self,bs.Actor.RETAINED,
                         (str(self._score) if self._scoreType == 'points' else bsUtils.getTimeString(self._score*10)),
                         maxWidth=300,
                         flash=True,trail=True,
                         scale = 1.0 if self._scoreType == 'points' else 0.6,
                         hAlign='center',
                         tiltTranslate=0.11,
                         position=(190+offsX,115),
                         jitter=1.0)
        bsUtils.Text(self,bs.Actor.RETAINED,
                     bs.getResource('finalScoreText')+':' if self._scoreType == 'points' else bs.getResource('finalTimeText')+':',
                     maxWidth=300,
                     position=(0,200),
                     transition='fadeIn',
                     hAlign='center',
                     vAlign='center',
                     transitionDelay=0)
        bs.gameTimer(350,bs.Call(bs.playSound,self.scoreDisplaySound))



class CoopGameActivity(bs.GameActivity):
    """
    category: Game Flow Classes

    Base class for cooperative-mode games.
    """

    @classmethod
    def supportsSessionType(cls,sessionType):
        return True if issubclass(sessionType,bs.CoopSession) else False

    def onBegin(self,nameSound=None):
        bs.GameActivity.onBegin(self,nameSound)

        # show achievements remaining..
        bs.gameTimer(3800,bs.WeakCall(self._showRemainingAchievements))

        # ..and preload their images in case we get some
        bs.gameTimer(2000,bs.WeakCall(self._preloadAchievements))

        # ..also lets ask the server for a 'time-to-beat' value
        levelName = self._getCoopLevelName()
        configStr = str(len(self.players))+"p"+self.getSession()._campaign.getLevelScoreVersion(self.settings['name']).replace(' ','_')
        bsInternal._getScoresToBeat(levelName,configStr,bs.WeakCall(self._onGotScoresToBeat))

    def _onGotScoresToBeat(self,score):
        pass

    def _showStandardScoresToBeatUI(self,scores):
        displayType = self._getScoreType()
        if scores is not None:
            # sort by originating date so that the most recent is first
            scores.sort(reverse=True,key=lambda score:score['time'])
            # now make a display for the most recent challenge
            for score in scores:
                if score['type'] == 'scoreChallenge':
                    t = bs.NodeActor(self,bs.Actor.RETAINED,
                                     bs.newNode('text',
                                                name='Score To Beat Text',
                                                owner=self,
                                                attrs={'vAttach':'top',
                                                       'hAttach':'center' if displayType == 'time' else 'left',
                                                       'hAlign':'center' if displayType == 'time' else 'left',
                                                       'color':(0.7,0.4,1,1),
                                                       'translate':(20,-70) if displayType == 'time' else (20,-130),
                                                       'scale':[0.6],
                                                       'text':score['player']+':  '+(bsUtils.getTimeString(int(score['value'])*10) if displayType == 'time' else str(score['value']))}))
                    
                    bsUtils.animate(t.node,'scale',{1000:0.0,1100:0.7,1200:0.6})
                    break

    # FIXME: need to expand on this to include stuff like ascending/descending
    # and should make it a public classmethod
    def _getScoreType(self):
        return 'points'

    def _getCoopLevelName(self):
        return self.getSession()._campaign.getName()+":"+str(self.settings['name'])

    def celebrate(self,duration):
        """
        Tells all existing player-controlled characters to celebrate;
        generally used when the good guys score or complete a wave.
        """
        for player in self.players:
            try: player.actor.node.handleMessage('celebrate',duration)
            except Exception: pass

    def _preloadAchievements(self):
        achievements = bsAchievement.getAchievementsForCoopLevel(self._getCoopLevelName())
        for a in achievements:
            a.getIconTexture(True)
        
    def _showRemainingAchievements(self):
        #objs = self._remainingAchievementsObjs = []
        tsHeight = -450
        tsHOffs = 30
        vOffs = -200
        achievements = [a for a in bsAchievement.getAchievementsForCoopLevel(self._getCoopLevelName()) if not a.isComplete()]

        if len(achievements) > 0:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('achievementsRemainingText'),
                         position=(tsHOffs-10,vOffs-10),
                         transition='fadeIn',
                         scale=1.1,
                         hAttach="left",
                         vAttach="top",
                         color=(0.8,0.8,1.0,1.0),
                         transitionDelay=0,
                         transitionOutDelay=1300 if self._isSlowMotion else 4000)
            h = 70
            v = -50
            tDelay = 0
            for a in achievements:
                tDelay += 50
                a.createDisplay(self,h,v+vOffs,0+tDelay,inGame=True,outDelay=1300 if self._isSlowMotion else 4000)
                v -= 55

            #bs.gameTimer(2000 if self._isSlowMotion else 5000,bs.WeakCall(self._killRemainingAchievementsObjs))

    # def _killRemainingAchievementsObjs(self):
    #     self._remainingAchievementsObjs = []

    def onPlayerLeaving(self,player):
        bs.GameActivity.onPlayerLeaving(self,player)
        bs.realTimer(2000,bs.WeakCall(self._endSessionIfEmpty))

    def _endSessionIfEmpty(self):
        # hmm shouldnt need this try/except here i dont think..
        try: players = self.players
        except Exception,e:
            print 'EXC bsCoopGame _endSessionIfEmpty cant get self.players; shouldnt happen:',e
            players = []
        if len(players) == 0:
            self.getSession().end()
        
    def spawnPlayerSpaz(self,player,position,angle=None):
        "Spawn and wire up a standard player spaz"

        spaz = bs.GameActivity.spawnPlayerSpaz(self,player,position,angle)

        # deaths are noteworthy in co-op games
        spaz.playBigDeathSound = True
        return spaz
        
    def _awardAchievement(self,achievement,sound=True):
        'award an achievement; returns True if a banner will be shown; False otherwise'
        # cache these for efficiency
        if not hasattr(self,'_achievementsAwarded'):
            self._achievementsAwarded = set()
        if achievement in self._achievementsAwarded:
            return False
        if not achievement in self._achievementsAwarded:
            self._achievementsAwarded.add(achievement)
            a =  bsAchievement.getAchievement(achievement)
            complete = a.isComplete()
            a.setComplete(sound=sound)
            return (not complete)

    def fadeToRed(self):
        """
        Fades the screen to red; useful when the good guys have lost.
        """
        cExisting = bs.getGlobals().tint
        c = bs.newNode("combine",owner=self,attrs={'input0':cExisting[0],'input1':cExisting[1],'input2':cExisting[2],'size':3})
        bs.animate(c,'input1',{0:cExisting[1],2000:0})
        bs.animate(c,'input2',{0:cExisting[2],2000:0})
        c.connectAttr('output',bs.getGlobals(),'tint')

    def getInstanceScoreBoardName(self):
        # use campaign name
        return self.getSession()._campaign.getLevelDisplayName(self.settings['name'])

    def setupLowLifeWarningSound(self):
        """
        Sets up the activity to play a beeping noise
        whenever any players are close to death.
        """
        
        self._lifeWarningBeep = None
        self._lifeWarningBeepTimer = bs.Timer(1000,bs.WeakCall(self._updateLifeWarning),repeat=True)
                
    def _updateLifeWarning(self):
        # beep continuously if anyone is close to death
        shouldBeep = False
        for player in self.players:
            if player.isAlive():
                # fixme - should abstract this instead of reading hitPoints directly
                if getattr(player.actor,'hitPoints',999) < 200:
                    shouldBeep = True
                    break
        if shouldBeep and self._lifeWarningBeep is None:
            try: warnBeepsSound = self._warnBeepsSound
            except Exception: warnBeepsSound = self._warnBeepsSound = bs.getSound('warnBeeps')
            self._lifeWarningBeep = bs.NodeActor(self,bs.Actor.UNRETAINED,bs.newNode('sound',owner=self,attrs={'sound':warnBeepsSound,'positional':False,'loop':True}))
        if self._lifeWarningBeep is not None and not shouldBeep:
            self._lifeWarningBeep = None

            
class CoopJoiningActivity(bsGame.JoiningActivity):
    def __init__(self,session,settings={}):
        bsGame.JoiningActivity.__init__(self,session,settings)

        # lets show a list of scores-to-beat for 1 player at least..
        levelName = session._campaign.getName()+":"+session._campaignInfo['level']
        configStr = "1p"+session._campaign.getLevelScoreVersion(session._campaignInfo['level']).replace(' ','_')
        bsInternal._getScoresToBeat(levelName,configStr,bs.WeakCall(self._onGotScoresToBeat))

    def onTransitionIn(self):
        bsGame.JoiningActivity.onTransitionIn(self)

        bsUtils.Text(self,bs.Actor.RETAINED,
                     self.getSession()._campaign.getLevelDisplayName(self.getSession()._campaignInfo['level']),
                     scale=1.3,
                     hAttach='center',
                     hAlign='center',
                     vAttach='top',
                     transition='fadeIn',
                     transitionDelay=400,
                     color=(1,1,1,0.6),
                     position=(0,-95))

        bs.ControlsHelpOverlay(self,bs.Actor.RETAINED,delay=1000)

    def _onGotScoresToBeat(self,scores):

        # sort by originating date so that the most recent is first
        if scores is not None:
            scores.sort(reverse=True,key=lambda score:score['time'])

        scoreType = self.getSession()._currentGameInstance._getScoreType()

        if scores is not None:
            achievementChallenges = [a for a in scores if a['type'] == 'achievementChallenge']
            scoreChallenges = [a for a in scores if a['type'] == 'scoreChallenge']
        else: achievementChallenges = scoreChallenges = []

        delay = 1000
        vpos = -140
        spacing = 25
        delayInc = 100

        def _addT(text,hOffs=0,scale=1.0,color=(1,1,1,0.46)):
            bsUtils.Text(self,bs.Actor.RETAINED,
                         text,scale=scale*0.76,hAlign='left',hAttach='left',vAttach='top',
                         transition='fadeIn',transitionDelay=delay,color=color,
                         position=(60+hOffs,vpos))

        if scoreChallenges:
            _addT(bs.getResource('scoreChallengesText')+':',scale=1.1)
            delay += delayInc
            vpos -= spacing
            for s in scoreChallenges:
                _addT(str(s['value'] if scoreType == 'points' else bsUtils.getTimeString(int(s['value'])*10))+'  (1 player)',hOffs=30,color=(0.9,0.7,1.0,0.8))
                delay += delayInc
                vpos -= 0.6*spacing
                _addT(s['player'],hOffs=40,color=(0.8,1,0.8,0.6),scale=0.8)
                delay += delayInc
                vpos -= 1.2*spacing
            vpos -= 0.5*spacing

        if achievementChallenges:
            _addT(bs.getResource('achievementChallengesText')+':',scale=1.1)
            delay += delayInc
            vpos -= spacing
            for s in achievementChallenges:
                _addT(str(s['value']),hOffs=30,color=(0.9,0.7,1.0,0.8))
                delay += delayInc
                vpos -= 0.6*spacing
                _addT(s['player'],hOffs=40,color=(0.8,1,0.8,0.6),scale=0.8)
                delay += delayInc
                vpos -= 1.2*spacing
            vpos -= 0.5*spacing

        # now list our remaining achievements for this level
        levelName = self.getSession()._campaign.getName()+":"+self.getSession()._campaignInfo['level']
        tsHeight = -450
        tsHOffs = 60
        achievements = [a for a in bsAchievement.getAchievementsForCoopLevel(levelName) if not a.isComplete()]

        bsUtils.Text(self,bs.Actor.RETAINED,
                     bs.getResource('achievementsRemainingText'),
                     position=(tsHOffs-10,vpos),
                     transition='fadeIn',
                     scale=1.1*0.76,
                     hAttach="left",
                     vAttach="top",
                     color=(1,1,1,0.5),
                     transitionDelay=delay)
        h = tsHOffs+50
        vpos -= 35
        for a in achievements:
            delay += 50
            a.createDisplay(self,h,vpos,delay,inGame=True,color=(1,1,1,0.5))
            vpos -= 55


        #if len(achievementChallenges) == 0 and len(scoreChallenges) == 0 and len(achievements) == 0:
        if len(achievements) == 0:
            bsUtils.Text(self,bs.Actor.RETAINED,
                         bs.getResource('noAchievementsRemainingText'),
                         position=(tsHOffs+15,vpos+10),
                         transition='fadeIn',
                         scale=0.7,
                         hAttach="left",
                         vAttach="top",
                         color=(1,1,1,0.5),
                         transitionDelay=delay+500)

        
class CoopSession(bs.Session):
    """
    category: Game Flow Classes

    A bs.Session which runs cooperative-mode games.
    These generally consist of 1-4 players against
    the computer and include functionality such as
    high score lists.
    """
    
    def __init__(self):
        """
        Instantiate a co-op mode session.
        """
        try: mp = bs.config['Coop Game Max Players']
        except Exception:
            # old pref value..
            try: mp = bs.config['Challenge Game Max Players']
            except Exception: mp = 4

        bs.Session.__init__(self,teamNames=gTeamNames,teamColors=gTeamColors,
                            useTeamColors=False,maxPlayers=mp)

        # pull some info on the game we're gonna play
        # (joining-activity displays info about this)
        self._campaignInfo = bsUI.gCampaignSelection
        self._campaign = getCampaign(self._campaignInfo['campaign'])

        self._ranTutorialActivity = False
        self._tutorialActivity = None
        
        # start our custom joining screen..
        j = CoopJoiningActivity(self)
        self.setActivity(j)

        self._updateOnDeckGameInstances()

        
    def _updateOnDeckGameInstances(self):

        # instantiates levels we might be running soon so they have time to load..
        
        # build an instance for the current level...
        gameTypeName = self._campaign.getLevelGameType(self._campaignInfo['level'])
        settings = self._campaign.getLevelSettings(self._campaignInfo['level'])
        moduleName,className = gameTypeName.split('.')
        module = __import__(moduleName)
        gameType = getattr(module,className)
        self._currentGameInstance =  gameType(self,settings)

        # find the next level and build an instance for it too...
        levels = self._campaign.getLevels()
        level = None
        for l in levels:
            if l['name'] == self._campaignInfo['level']:
                level = l
                break
        if level == None: raise Exception()
        if level['index'] < len(levels)-1: nextLevel = levels[level['index']+1]
        else: nextLevel = None
        if nextLevel:
            gameTypeName = self._campaign.getLevelGameType(nextLevel['name'])
            settings = self._campaign.getLevelSettings(nextLevel['name'])
            moduleName,className = gameTypeName.split('.')
            module = __import__(moduleName)
            gameType = getattr(module,className)
            self._nextGameInstance =  gameType(self,settings)
            self._nextGameName = nextLevel['name']
        else:
            self._nextGameInstance = None
            self._nextGameName = None

        # if our current level is 'onslaught training', instantiate our tutorial
        # so its ready to go.. (if we havnt run it yet)
        if (self._campaignInfo['level'] == 'Onslaught Training'
            and self._tutorialActivity is None
            and not self._ranTutorialActivity):
            self._tutorialActivity = bsTutorial.TutorialActivity(self)
        

    def onPlayerRequest(self,player):
        """
        Method override; currently rejects players trying to join mid-game.
        """
        if (isinstance(self._activity,bsGame.JoiningActivity)):
            return bs.Session.onPlayerRequest(self,player)
        else:
            bs.screenMessage(bs.getResource('noJoinCoopMidwayText'))
            bs.playSound(bs.getSound('error'))
            return False

    def restart(self):
        """
        Restarts the current game activity.
        """
        # tell the current activity to end with a 'restart' outcome
        # we use 'force' so that we apply even if end has already been called
        # (but is in its delay period)

        # make an exception if there's no players left.. otherwise this can override
        # the default session end that occurs in that case
        if len(self.players) < 1: return
        self._activity.end(results={'outcome':'restart'},force=True)

    def onActivityEnd(self,activity,results):
        """
        Method override for co-op sessions;
        jumps between co-op games and score screens.
        """
        try: outcome = results['outcome']
        except Exception: outcome = ''

        # if we're in a between-round activity or a restart-activity, hop into a round
        if (isinstance(activity,bsGame.JoiningActivity)
            or isinstance(activity,CoopScoreScreen)
            or isinstance(activity,bsGame.TransitionActivity)):

            if outcome == 'nextLevel':
                if self._nextGameInstance is None: raise Exception()
                self._campaignInfo['level'] = self._nextGameName
                nextGame = self._nextGameInstance
            else:
                nextGame = self._currentGameInstance

            # special case: if we're coming from a joining-activity and will be going into
            # onslaught-training, show the tutorial first..
            if isinstance(activity,bsGame.JoiningActivity) and self._campaignInfo['level'] == 'Onslaught Training':
                if self._tutorialActivity is None: raise Exception("_tutorialActivity not ready when it should be")
                self.setActivity(self._tutorialActivity)
                self._tutorialActivity = None
                self._ranTutorialActivity = True
                bsUI.customMenuUI = []

            # normal case; launch the next round
            else:

                # do a full reset on our score-set and point it at our new activity
                self.scoreSet.reset()
                for p in self.players: self.scoreSet.registerPlayer(p)
                self.scoreSet.setActivity(nextGame)

                # now flip the current activity
                self.setActivity(nextGame)

                # get new cached instances of the current and next level loading..
                #self._updateOnDeckGameInstances()

                # fixme; should make an official call for this
                bsUI.customMenuUI = [[bs.getResource('restartText'),bs.WeakCall(self.restart)]]

        # if we were in a tutorial, just pop a transition to get to the actual round
        elif isinstance(activity,bsTutorial.TutorialActivity):
            self.setActivity(bsGame.TransitionActivity(self))
        else:

            # looks like we were in a round - check the outcome and go from there
            if outcome == 'restart':
                self.setActivity(bsGame.TransitionActivity(self)) # this will pop up back in the same round
            else:
                self.setActivity(CoopScoreScreen(self,{'playerInfo':results['playerInfo'],
                                                       'score':results['score'] if 'score' in results else None,
                                                       'failMessage':results['failMessage'] if 'failMessage' in results else None,
                                                       'scoreOrder':results['scoreOrder'] if 'scoreOrder' in results else 'increasing',
                                                       'scoreType':activity._getScoreType(),
                                                       'outcome':outcome,
                                                       'campaign':self._campaign,
                                                       'level':self._campaignInfo['level']}))

        # no matter what, get the next 2 levels ready to go..
        self._updateOnDeckGameInstances()
                

# fill out our campaigns
c = Campaign('Default')
c.addLevel('Onslaught Training','bsOnslaught.OnslaughtGame',config={'preset':'training'})
c.addLevel('Rookie Onslaught','bsOnslaught.OnslaughtGame',config={'preset':'rookie'})
c.addLevel('Rookie Football','bsFootball.FootballCoopGame',config={'preset':'rookie'},scoreVersion=' B')
c.addLevel('Pro Onslaught','bsOnslaught.OnslaughtGame',config={'preset':'pro'})
c.addLevel('Pro Football','bsFootball.FootballCoopGame',config={'preset':'pro'},scoreVersion=' B')
c.addLevel('Pro Runaround','bsRunaround.RunaroundGame',config={'preset':'pro'})
c.addLevel('Uber Onslaught','bsOnslaught.OnslaughtGame',config={'preset':'uber'})
c.addLevel('Uber Football','bsFootball.FootballCoopGame',config={'preset':'uber'},scoreVersion=' B')
c.addLevel('Uber Runaround','bsRunaround.RunaroundGame',config={'preset':'uber'})
c.addLevel('The Last Stand','bsTheLastStand.TheLastStandGame',config={})
_registerCampaign(c)


c = Campaign('Challenges',sequential=False)
c.addLevel('Onslaught','bsOnslaught.OnslaughtGame',config={'preset':'endless'})
c.addLevel('Runaround','bsRunaround.RunaroundGame',config={'preset':'endless'})
_registerCampaign(c)

# def getCoopCampaigns():
#     return [
#         {'name':'Default',
#          'levels':[
#                 {'name':'Onslaught Training',
#                  'type':'bsOnslaught.OnslaughtGame',
#                  'settings':{
#                         'preset':'training'
#                         }
#                 }
#                 ]
#         ]
