import bs
import bsUtils
import random
import bsVector

_maps = {}

def preloadPreviewMedia():
    bs.getModel('levelSelectButtonOpaque')
    bs.getModel('levelSelectButtonTransparent')
    for m in _maps.values():
        mapTexName = m.getPreviewTextureName()
        if mapTexName is not None: bs.getTexture(mapTexName)
    
def registerMap(m):
    """ Register a map class with the game. """
    if _maps.has_key(m.name):
        raise Exception("map \"" + m.name + "\" already registered")
    _maps[m.name] = m

def getFilteredMapName(name):
    """ filters a map name to account for name changes, etc so old configs still work """
    # some legacy name fallbacks...
    if name == 'AlwaysLand' or name == 'Happy Land': name = 'Happy Thoughts'
    if name == 'Hockey Arena': name = 'Hockey Stadium'
    return name

def getTranslatedMapName(name):
    try: translated = bs.getResource('translations.mapsNames')[name]
    except Exception: translated = None
    if translated is None: translated = name
    return translated
    
def getMapsSupportingPlayType(playType):
    """
    category: Media Functions

    Return a list of bs.Map types supporting a specified play-type (a string).
    Maps supporting a given play-type must provide a particular set of
    features or lend themselves to a certain style of play.

    Play Types:
    
    'melee' - general fighting map - has 2+ 'spawn' pts, 1+ 'powerupSpawn' pts

    'teamFlag' - for CTF, etc - has 2+ 'spawn' pts, 2+ 'flag' pts, and 1+ 'powerupSpawn' pts

    'keepAway'- has 2+ 'spawn' pts, 1+ 'flagDefault' pts, and 1+ 'powerupSpawn' pts

    'conquest' - has 2+ 'flag' pts, 2+ 'spawnByFlag' pts, and 1+ 'powerupSpawn' pts

    'kingOfTheHill' - has 2+ 'spawn' pts, 1+ 'flagDefault' pts, and 1+ 'powerupSpawn' pts

    'hockey' - has 2 'goal' pts, 2+ 'spawn' pts, 1+ 'flagDefault' pts, 1+ 'powerupSpawn' pts

    'football' - has 2 'goal' pts, 2+ 'spawn' pts, 1+ 'flagDefault' pts, 1+ 'powerupSpawn' pts
    
    'race' - has 2+ 'racePoint' pts
    """
    maps = [m[0] for m in _maps.items() if playType in m[1].playTypes]
    maps.sort()
    return maps

def getMapClass(name):
    """ return a map type given a name """
    name = getFilteredMapName(name)
    try: return _maps[name]
    except Exception: raise Exception("Map not found: '"+name+"'")
    
class Map(bs.Actor):
    """
    category: Game Flow Classes

    A collection of terrain nodes, metadata, and other
    functionality comprising a game map.
    """
    defs = None
    name = "Map"
    playTypes = []
    preloaded = False

    @classmethod
    def preload(cls):
        """ Preload map media.
        This runs the class's onPreload function if need be to get it ready to run.
        Preloading can be fired for a soon-to-be-needed map to speed its creation.
        This is a classmethod since it is not run on map instances but rather on
        the class as a whole before instances are made"""
        if not cls.preloaded:
            cls.onPreload()
            cls.preloaded = True

    @classmethod
    def getPreviewTextureName(cls):
        """
        Return the name of the preview texture for this map.
        """
        return None

    @classmethod
    def onPreload(cls):
        """
        Called when the map is being preloaded;
        it should load any media it requires to
        class attributes on itself.
        """
        pass

    @classmethod
    def getName(cls):
        """
        Return the unique name of this map, in English.
        """
        return cls.name

    @classmethod
    def getMusicType(cls):
        """
        Returns a particular music-type string that should be played on
        this map; or None if the default music should be used.
        """
        return None

    def __init__(self,activity,retain):
        """
        Instantiate a map.
        """
        import bsInternal
        bs.Actor.__init__(self,activity,retain)
        self.preload()

        # set some defaults
        bsGlobals = bs.getGlobals()
        bsGlobals.happyThoughts = 0
        bsGlobals.floorReflection = False
        bsGlobals.debrisFriction = 1.0
        bsGlobals.debrisKillHeight = -50.0
        bsGlobals.areaOfInterestBounds = self.getDefBoundBox("areaOfInterestBounds") or (-500,-500,-500,500,500,500)
        bsGlobals.shadowOffset = (0,0,0)
        bsGlobals.shadowScale = (1.0,1.0)
        bsGlobals.shadowOrtho = False
        bsGlobals.tint = (1.1,1.0,0.9)
        bsGlobals.ambientColor = (1,1,1)
        bsGlobals.vignetteOuter = (0.6,0.6,0.6)
        bsGlobals.vignetteInner = (0.95,0.95,0.95)

        bsInternal._setMapBounds(self.getDefBoundBox("levelBounds") or (-30,-10,-30,30,100,30))

        try: bsGlobals.shadowRange = [self.defs.points[v][1] for v in ['shadowLowerBottom','shadowLowerTop','shadowUpperBottom','shadowUpperTop']]
        except Exception: bsGlobals.shadowRange = [-4,0,10,15]

        self.spawnPoints = self.getDefPoints("spawn") or [(0,0,0,0,0,0)]
        self.ffaSpawnPoints = self.getDefPoints("ffaSpawn") or [(0,0,0,0,0,0)]
        self.spawnByFlagPoints = self.getDefPoints("spawnByFlag") or [(0,0,0,0,0,0)]
        self.flagPoints = self.getDefPoints("flag") or [(0,0,0)]
        self.flagPointDefault = self.getDefPoint("flagDefault") or (0,1,0)
        self.powerupSpawnPoints = self.getDefPoints("powerupSpawn") or [(0,0,0)]
        self.tntPoints = self.getDefPoints("tnt") or []
        self.isHockey = False
        self.isFlying = False
        self._nextFFAStartIndex = 0

    def _isPointNearEdge(self,p,running=False):
        "For bot purposes.."
        return False

    def getDefBoundBox(self,name):
        """Returns a 6 member bounds tuple or None if it is not defined."""
        try:
            b = self.defs.boxes[name]
            return (b[0]-b[6]/2.0,b[1]-b[7]/2.0,b[2]-b[8]/2.0,
                    b[0]+b[6]/2.0,b[1]+b[7]/2.0,b[2]+b[8]/2.0);
        except Exception:
            return None
        
    def getDefPoint(self,name):
        """Returns a single defined point or a default value in its absence."""
        try:
            return self.defs.points[name]
        except Exception:
            return None

    def getDefPoints(self,name):
        """
        Returns a list of points - as many sequential ones are defined
        (flag1, flag2, flag3), etc.
        """
        if self.defs and self.defs.points.has_key(name+"1"):
            pointList = []
            i = 1
            while self.defs.points.has_key(name+str(i)):
                p = self.defs.points[name+str(i)]
                if len(p) == 6:
                    pointList.append(p)
                else:
                    if len(p) != 3: raise Exception("invalid point")
                    pointList.append(p+(0,0,0))
                i += 1
            return pointList
        else:
            return None
        
    def getStartPosition(self,teamIndex):
        """
        Returns a random starting position in the map for the given team index.
        """
        pt = self.spawnPoints[teamIndex%len(self.spawnPoints)]
        xRange = (-0.5,0.5) if pt[3] == 0 else (-pt[3],pt[3])
        zRange = (-0.5,0.5) if pt[5] == 0 else (-pt[5],pt[5])
        pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
        return pt

    def getFFAStartPosition(self,players):
        """
        Returns a random starting position in one of the FFA spawn areas.
        If a list of bs.Players is provided; the returned points will be
        as far from these players as possible.
        """

        # get positions for existing players
        playerPts = []
        for player in players:
            try:
                pt = bsVector.Vector(*player.actor.node.translate)
                # fresh players will have translate 0,0,0; need to fix this..
                if pt.gameData == [0.0,0.0,0.0]:
                    pt = bsVector.Vector(*player.actor._lastStandPos)
                playerPts.append(pt)
            except Exception: pass

        def _getPt():
            pt = self.ffaSpawnPoints[self._nextFFAStartIndex]
            self._nextFFAStartIndex = (self._nextFFAStartIndex+1)%len(self.ffaSpawnPoints)
            xRange = (-0.5,0.5) if pt[3] == 0 else (-pt[3],pt[3])
            zRange = (-0.5,0.5) if pt[5] == 0 else (-pt[5],pt[5])
            pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
            return pt

        if len(playerPts) == 0:
            return _getPt()
        else:
            # lets calc several start points and then pick whichever is farthest from all existing players
            farthestPtDist = -1.0
            farthestPt = None
            for i in range(10):
                testPt = bsVector.Vector(*_getPt())
                closestPlayerDist = 9999.0
                closestPlayerPt = None
                for pp in playerPts:
                    dist = (pp-testPt).length()
                    if dist < closestPlayerDist:
                        closestPlayerDist = dist
                        closestPlayerPt = pp
                if closestPlayerDist > farthestPtDist:
                    farthestPtDist = closestPlayerDist
                    farthestPt = testPt
            return tuple(farthestPt.data)

    def getFlagPosition(self,teamIndex):
        """
        Return a flag position on the map for the given team index.
        Pass None to get the default flag point.
        (used for things such as king-of-the-hill)
        """
        if teamIndex is None:
            return self.flagPointDefault[:3]
        else:
            return self.flagPoints[teamIndex%len(self.flagPoints)][:3]

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage): self.node.delete()
        else: bs.Actor.handleMessage(self,m)



######### now lets go ahead and register some maps #########


class HockeyStadium(Map):
    import hockeyStadiumDefs as defs
    name = "Hockey Stadium"

    playTypes = ['melee','hockey','teamFlag','keepAway']

    @classmethod
    def getPreviewTextureName(cls):
        return 'hockeyStadiumPreview'

    @classmethod
    def onPreload(cls):
        cls.models = (bs.getModel("hockeyStadiumOuter"),
                            bs.getModel("hockeyStadiumInner"),
                            bs.getModel("hockeyStadiumStands"))
        cls.collideModel = bs.getCollideModel("hockeyStadiumCollide")
        cls.tex = bs.getTexture("hockeyStadium")
        cls.standsTex = bs.getTexture("footballStadium")
        cls.iceMaterial = bs.newMaterial()
        cls.iceMaterial.addActions(actions=("modifyPartCollision","friction",0.01))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={'model':self.models[0],
                                      'collideModel':self.collideModel,
                                      'colorTexture':self.tex,
                                      'materials':[bs.footingMaterial,self.iceMaterial]})
        self.floor = bs.newNode("terrain",
                                owner=self.node,
                                attrs={"model":self.models[1],
                                       "colorTexture":self.tex,
                                       "opacity":0.92,
                                       "opacityInLowOrMediumQuality":1.0,
                                       "materials":[bs.footingMaterial,self.iceMaterial]})
        self.stands = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.models[2],
                                        "visibleInReflections":False,
                                        "colorTexture":self.standsTex})
        bsGlobals = bs.getGlobals()
        bsGlobals.floorReflection = True
        bsGlobals.debrisFriction = 0.3
        bsGlobals.debrisKillHeight = -0.3

        bsGlobals.tint = (1.2,1.3,1.33)
        bsGlobals.ambientColor = (1.15,1.25,1.6)
        bsGlobals.vignetteOuter = (0.66,0.67,0.73)
        bsGlobals.vignetteInner = (0.93,0.93,0.95)

        self.isHockey = True

registerMap(HockeyStadium)


class FootballStadium(Map):
    import footballStadiumDefs as defs
    name = "Football Stadium"

    playTypes = ['melee','football','teamFlag','keepAway']

    @classmethod
    def getPreviewTextureName(cls):
        return 'footballStadiumPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("footballStadium")
        cls.collideModel = bs.getCollideModel("footballStadiumCollide")
        cls.tex = bs.getTexture("footballStadium")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"model":self.model,
                                      "collideModel":self.collideModel,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.3,1.2,1.0)
        bsGlobals.ambientColor = (1.3,1.2,1.0)
        bsGlobals.vignetteOuter = (0.57,0.57,0.57)
        bsGlobals.vignetteInner = (0.9,0.9,0.9)

    def _isPointNearEdge(self,p,running=False):
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(FootballStadium)


class BridgitMap(Map):
    import bridgitLevelDefs as defs
    name = "Bridgit"
    playTypes = ["melee","teamFlag",'keepAway']

    @classmethod
    def getPreviewTextureName(cls):
        return 'bridgitPreview'

    @classmethod
    def onPreload(cls):
        cls.modelTop = bs.getModel("bridgitLevelTop")
        cls.modelBottom = bs.getModel("bridgitLevelBottom")
        cls.modelBG = bs.getModel("natureBackground")
        cls.collideModel = bs.getCollideModel("bridgitLevelCollide")
        cls.tex = bs.getTexture("bridgitLevelColor")
        cls.modelBGTex = bs.getTexture("natureBackgroundColor")
        cls.collideBG = bs.getCollideModel("natureBackgroundCollide")

        cls.railingCollideModel = bs.getCollideModel("bridgitLevelRailingCollide")
    
        cls.bgMaterial = bs.newMaterial()
        cls.bgMaterial.addActions(actions=('modifyPartCollision','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.modelTop,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.1,1.2,1.3)
        bsGlobals.ambientColor = (1.1,1.2,1.3)
        bsGlobals.vignetteOuter = (0.65,0.6,0.55)
        bsGlobals.vignetteInner = (0.9,0.9,0.93)

registerMap(BridgitMap)


class BigGMap(Map):
    import bigGDefs as defs
    name = "Big G"
    playTypes = ["race",'melee','keepAway','teamFlag','kingOfTheHill','conquest']

    @classmethod
    def getPreviewTextureName(cls):
        return 'bigGPreview'

    @classmethod
    def onPreload(cls):
        cls.modelTop = bs.getModel("bigG")
        cls.modelBottom = bs.getModel("bigGBottom")
        cls.modelBG = bs.getModel("natureBackground")
        cls.collideModel = bs.getCollideModel("bigGCollide")
        cls.tex = bs.getTexture("bigG")
        cls.modelBGTex = bs.getTexture("natureBackgroundColor")
        cls.collideBG = bs.getCollideModel("natureBackgroundCollide")
        cls.bumperCollideModel = bs.getCollideModel("bigGBumper")

        cls.bgMaterial = bs.newMaterial()
        cls.bgMaterial.addActions(actions=('modifyPartCollision','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "color":(0.7,0.7,0.7),
                                      "model":self.modelTop,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "color":(0.7,0.7,0.7),
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.bumperCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.1,1.2,1.3)
        bsGlobals.ambientColor = (1.1,1.2,1.3)
        bsGlobals.vignetteOuter = (0.65,0.6,0.55)
        bsGlobals.vignetteInner = (0.9,0.9,0.93)

registerMap(BigGMap)


class RoundaboutMap(Map):
    import roundaboutLevelDefs as defs
    name = "Roundabout"
    playTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'roundaboutPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("roundaboutLevel")
        cls.modelBottom = bs.getModel("roundaboutLevelBottom")
        cls.modelBG = bs.getModel("natureBackground")
        cls.collideModel = bs.getCollideModel("roundaboutLevelCollide")
        cls.tex = bs.getTexture("roundaboutLevelColor")
        cls.modelBGTex = bs.getTexture("natureBackgroundColor")
        cls.collideBG = bs.getCollideModel("natureBackgroundCollide")
    
        cls.railingCollideModel = bs.getCollideModel("roundaboutLevelBumper")

        cls.bgMaterial = bs.newMaterial()
        cls.bgMaterial.addActions(actions=('modifyPartCollision','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.modelBG,
                                    "lighting":False,
                                    "colorTexture":self.modelBGTex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.0,1.05,1.1)
        bsGlobals.ambientColor = (1.0,1.05,1.1)
        bsGlobals.shadowOrtho = True
        bsGlobals.vignetteOuter = (0.63,0.65,0.7)
        bsGlobals.vignetteInner = (0.97,0.95,0.93)

registerMap(RoundaboutMap)


class MonkeyFaceMap(Map):
    import monkeyFaceLevelDefs as defs
    name = "Monkey Face"
    playTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'monkeyFacePreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("monkeyFaceLevel")
        cls.bottomModel = bs.getModel("monkeyFaceLevelBottom")
        cls.modelBG = bs.getModel("natureBackground")
        cls.collideModel = bs.getCollideModel("monkeyFaceLevelCollide")
        cls.tex = bs.getTexture("monkeyFaceLevelColor")
        cls.modelBGTex = bs.getTexture("natureBackgroundColor")
        cls.collideBG = bs.getCollideModel("natureBackgroundCollide")
    
        cls.railingCollideModel = bs.getCollideModel("monkeyFaceLevelBumper")

        cls.bgMaterial = bs.newMaterial()
        cls.bgMaterial.addActions(actions=('modifyPartCollision','friction',10.0))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.1,1.2,1.2)
        bsGlobals.ambientColor = (1.2,1.3,1.3)
        bsGlobals.vignetteOuter = (0.60,0.62,0.66)
        bsGlobals.vignetteInner = (0.97,0.95,0.93)

registerMap(MonkeyFaceMap)


class ZigZagMap(Map):
    import zigZagLevelDefs as defs
    name = "Zigzag"
    playTypes = ["melee","keepAway","teamFlag","conquest",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(cls):
        return 'zigzagPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("zigZagLevel")
        cls.modelBottom = bs.getModel("zigZagLevelBottom")
        cls.modelBG = bs.getModel("natureBackground")
        cls.collideModel = bs.getCollideModel("zigZagLevelCollide")
        cls.tex = bs.getTexture("zigZagLevelColor")
        cls.modelBGTex = bs.getTexture("natureBackgroundColor")
        cls.collideBG = bs.getCollideModel("natureBackgroundCollide")

        cls.railingCollideModel = bs.getCollideModel("zigZagLevelBumper")

        cls.bgMaterial = bs.newMaterial()
        cls.bgMaterial.addActions(actions=('modifyPartCollision','friction',10.0))
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.modelBG,
                                     "lighting":False,
                                     "colorTexture":self.modelBGTex})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.modelBottom,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,self.bgMaterial,bs.deathMaterial]})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.0,1.15,1.15)
        bsGlobals.ambientColor = (1.0,1.15,1.15)
        bsGlobals.vignetteOuter = (0.57,0.59,0.63)
        bsGlobals.vignetteInner = (0.97,0.95,0.93)

registerMap(ZigZagMap)


class ThePadMap(Map):
    import thePadLevelDefs as defs
    name = "The Pad"
    playTypes = ["melee","keepAway","teamFlag",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(cls):
        return 'thePadPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("thePadLevel")
        cls.bottomModel = bs.getModel("thePadLevelBottom")
        cls.collideModel = bs.getCollideModel("thePadLevelCollide")
        cls.tex = bs.getTexture("thePadLevelColor")
        cls.bgTex = bs.getTexture("menuBG")
        cls.bgModel = bs.getModel("thePadBGSmall")
        cls.railingCollideModel = bs.getCollideModel("thePadLevelBumper")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.1,1.1,1.0)
        bsGlobals.ambientColor = (1.1,1.1,1.0)
        bsGlobals.vignetteOuter = (0.7,0.65,0.75)
        bsGlobals.vignetteInner = (0.95,0.95,0.93)

registerMap(ThePadMap)


class DoomShroomMap(Map):
    import doomShroomLevelDefs as defs
    name = "Doom Shroom"
    playTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'doomShroomPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("doomShroomLevel")
        cls.collideModel = bs.getCollideModel("doomShroomLevelCollide")
        cls.tex = bs.getTexture("doomShroomLevelColor")
        cls.bgTex = bs.getTexture("doomShroomBGColor")
        cls.bgModel = bs.getModel("doomShroomBG")
        cls.stemModel = bs.getModel("doomShroomStem")
        cls.collideBG = bs.getCollideModel("doomShroomStemCollide")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               name='Terrain Root',
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex},
                              name='Terrain BG')
        self.stem = bs.newNode("terrain",
                               owner=self.node,
                               attrs={"model":self.stemModel,
                                      "lighting":False,
                                      "colorTexture":self.tex},
                               name='Terrain Unlit')
        self.bgCollide = bs.newNode("terrain",
                                    owner=self.node,
                                    attrs={"collideModel":self.collideBG,
                                           "materials":[bs.footingMaterial,bs.deathMaterial]},
                                    name='Terrain BG Collide')
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (0.82,1.10,1.15)
        bsGlobals.ambientColor = (0.9,1.3,1.1)
        bsGlobals.shadowOrtho = False
        bsGlobals.vignetteOuter = (0.76,0.76,0.76)
        bsGlobals.vignetteInner = (0.95,0.95,0.99)

    def _isPointNearEdge(self,p,running=False):
        x = p.x()
        z = p.z()
        xAdj = x*0.125
        zAdj = (z+3.7)*0.2
        if running:
            xAdj *= 1.4
            zAdj *= 1.4
        return (xAdj*xAdj+zAdj*zAdj > 1.0)

registerMap(DoomShroomMap)


class TipTopMap(Map):
    import tipTopLevelDefs as defs
    name = "Tip Top"
    playTypes = ["melee","keepAway","teamFlag",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(cls):
        return 'tipTopPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("tipTopLevel")
        cls.bottomModel = bs.getModel("tipTopLevelBottom")
        cls.collideModel = bs.getCollideModel("tipTopLevelCollide")
        cls.tex = bs.getTexture("tipTopLevelColor")
        cls.bgTex = bs.getTexture("tipTopBGColor")
        cls.bgModel = bs.getModel("tipTopBG")
        cls.railingCollideModel = bs.getCollideModel("tipTopLevelBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "color":(0.7,0.7,0.7),
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "color":(0.7,0.7,0.7),
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "color":(0.4,0.4,0.4),
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (0.8,0.9,1.3)
        bsGlobals.ambientColor = (0.8,0.9,1.3)
        bsGlobals.vignetteOuter = (0.79,0.79,0.69)
        bsGlobals.vignetteInner = (0.97,0.97,0.99)

registerMap(TipTopMap)


class CragCastleMap(Map):
    import cragCastleDefs as defs
    name = "Crag Castle"
    playTypes = ["melee","keepAway","teamFlag",'conquest']

    @classmethod
    def getPreviewTextureName(cls):
        return 'cragCastlePreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("cragCastleLevel")
        cls.bottomModel = bs.getModel("cragCastleLevelBottom")
        cls.collideModel = bs.getCollideModel("cragCastleLevelCollide")
        cls.tex = bs.getTexture("cragCastleLevelColor")
        cls.bgTex = bs.getTexture("menuBG")
        cls.bgModel = bs.getModel("thePadBG")
        cls.railingCollideModel = bs.getCollideModel("cragCastleLevelBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.shadowOrtho = True
        bsGlobals.shadowOffset = (0,0,-5.0)
        bsGlobals.tint = (1.15,1.05,0.75)
        bsGlobals.ambientColor = (1.15,1.05,0.75)
        bsGlobals.vignetteOuter = (0.6,0.65,0.6)
        bsGlobals.vignetteInner = (0.95,0.95,0.95)

registerMap(CragCastleMap)



class TowerDMap(Map):
    import towerDLevelDefs as defs
    name = "Tower D"
    playTypes = []

    @classmethod
    def getPreviewTextureName(cls):
        return 'towerDPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("towerDLevel")
        cls.collideModel = bs.getCollideModel("towerDLevelCollide")
        cls.tex = bs.getTexture("towerDLevelColor")
        cls.bgTex = bs.getTexture("menuBG")
        cls.bgModel = bs.getModel("thePadBG")
        cls.playerWallCollideModel = bs.getCollideModel("towerDPlayerWall")
        cls.playerWallMaterial = bs.newMaterial()
        cls.playerWallMaterial.addActions(actions=(("modifyPartCollision","friction",0.0)))

        # anything that needs to hit the wall should apply this.
        cls.collideWithWallMaterial = bs.newMaterial()
        cls.playerWallMaterial.addActions(
            conditions=("theyDontHaveMaterial",cls.collideWithWallMaterial),
            actions=("modifyPartCollision","collide",False))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.playerWall = bs.newNode("terrain",
                                     owner=self.node,
                                     attrs={"collideModel":self.playerWallCollideModel,
                                            "affectBGDynamics":False,
                                            "materials":[self.playerWallMaterial]})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.15,1.11,1.03)
        bsGlobals.ambientColor = (1.2,1.1,1.0)
        bsGlobals.vignetteOuter = (0.7,0.73,0.7)
        bsGlobals.vignetteInner = (0.95,0.95,0.95)

    def _isPointNearEdge(self,p,running=False):
        # see if we're within edgeBox
        boxes = self.defs.boxes
        boxTranslate = boxes['edgeBox'][0:3]
        boxScale = boxes['edgeBox'][6:9]
        boxTranslate2 = boxes['edgeBox2'][0:3]
        boxScale2 = boxes['edgeBox2'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        x2 = (p.x() - boxTranslate2[0])/boxScale2[0]
        z2 = (p.z() - boxTranslate2[2])/boxScale2[2]
        # if we're outside of *both* boxes we're near the edge
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5) and (x2 < -0.5 or x2 > 0.5 or z2 < -0.5 or z2 > 0.5)

registerMap(TowerDMap)


class AlwaysLandMap(Map):
    import alwaysLandLevelDefs as defs
    name = "Happy Thoughts"
    playTypes = ["melee","keepAway","teamFlag","conquest",'kingOfTheHill']

    @classmethod
    def getPreviewTextureName(cls):
        return 'alwaysLandPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("alwaysLandLevel")
        cls.bottomModel = bs.getModel("alwaysLandLevelBottom")
        cls.bgModel = bs.getModel("alwaysLandBG")
        cls.collideModel = bs.getCollideModel("alwaysLandLevelCollide")
        cls.tex = bs.getTexture("alwaysLandLevelColor")
        cls.bgTex = bs.getTexture("alwaysLandBGColor")

    @classmethod
    def getMusicType(cls):
        return 'Flying'
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.foo = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex})
        bsGlobals = bs.getGlobals()
        bsGlobals.happyThoughts = 1
        bsGlobals.shadowOffset = (0.0,8.0,5.0)
        bsGlobals.tint = (1.3,1.23,1.0)
        bsGlobals.ambientColor = (1.3,1.23,1.0)
        bsGlobals.vignetteOuter = (0.64,0.59,0.69)
        bsGlobals.vignetteInner = (0.95,0.95,0.93)
        self.isFlying = True

        # throw out some tips on flying
        t = bs.newNode('text',
                       owner=self.getActivity(),
                       attrs={'text':bs.getResource('pressJumpToFlyText'),
                              'scale':[1.1],
                              'translate':(0,200),
                              'hAlign':'center',
                              'vAttach':'bottom'},name='Tip')
        c = bs.newNode("combine",owner=t,attrs={'input0':0.8,'input1':1.0,'input2':1.0})
        c.connectAttr('output',t,'color')
        bsUtils.animate(c,'input3',{0:0,1000:1,7000:1,8000:0})
        bs.gameTimer(8000,t.delete)
        
registerMap(AlwaysLandMap)


class StepRightUpMap(Map):
    import stepRightUpLevelDefs as defs
    name = "Step Right Up"
    playTypes = ["melee","keepAway","teamFlag","conquest"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'stepRightUpPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("stepRightUpLevel")
        cls.collideModel = bs.getCollideModel("stepRightUpLevelCollide")
        cls.tex = bs.getTexture("stepRightUpLevelColor")
        cls.bgTex = bs.getTexture("menuBG")
        cls.bgModel = bs.getModel("thePadBG")

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.2,1.1,1.0)
        bsGlobals.ambientColor = (1.2,1.1,1.0)
        bsGlobals.vignetteOuter = (0.7,0.65,0.75)
        bsGlobals.vignetteInner = (0.95,0.95,0.93)

registerMap(StepRightUpMap)


class CourtyardMap(Map):
    import courtyardLevelDefs as defs
    name = "Courtyard"
    playTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'courtyardPreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("courtyardLevel")
        cls.collideModel = bs.getCollideModel("courtyardLevelCollide")
        cls.tex = bs.getTexture("courtyardLevelColor")
        cls.bgTex = bs.getTexture("menuBG")
        cls.bgModel = bs.getModel("thePadBG")
        cls.playerWallCollideModel = bs.getCollideModel("courtyardPlayerWall")
    
        cls.playerWallMaterial = bs.newMaterial()
        cls.playerWallMaterial.addActions(actions=(("modifyPartCollision","friction",0.0)))

        # anything that needs to hit the wall should apply this.
        cls.collideWithWallMaterial = bs.newMaterial()
        cls.playerWallMaterial.addActions(
            conditions=("theyDontHaveMaterial",cls.collideWithWallMaterial),
            actions=("modifyPartCollision","collide",False))

    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})

        # in challenge games, put up a wall to prevent players
        # from getting in the turrets (that would foil our brilliant AI)
        if 'CoopSession' in str(type(bs.getSession())):
            self.playerWall = bs.newNode("terrain",
                                         owner=self.node,
                                         attrs={"collideModel":self.playerWallCollideModel,
                                                "affectBGDynamics":False,
                                                "materials":[self.playerWallMaterial]})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.2,1.17,1.1)
        bsGlobals.ambientColor = (1.2,1.17,1.1)
        bsGlobals.vignetteOuter = (0.6,0.6,0.64)
        bsGlobals.vignetteInner = (0.95,0.95,0.93)

    def _isPointNearEdge(self,p,running=False):
        # count anything off our ground level as safe (for our platforms)
        #if p.y() > 3.1: return False
        # see if we're within edgeBox
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(CourtyardMap)


class RampageMap(Map):
    import rampageLevelDefs as defs
    name = "Rampage"
    playTypes = ["melee","keepAway","teamFlag"]

    @classmethod
    def getPreviewTextureName(cls):
        return 'rampagePreview'

    @classmethod
    def onPreload(cls):
        cls.model = bs.getModel("rampageLevel")
        cls.bottomModel = bs.getModel("rampageLevelBottom")
        cls.collideModel = bs.getCollideModel("rampageLevelCollide")
        cls.tex = bs.getTexture("rampageLevelColor")
        cls.bgTex = bs.getTexture("rampageBGColor")
        cls.bgTex2 = bs.getTexture("rampageBGColor2")
        cls.bgModel = bs.getModel("rampageBG")
        cls.bgModel2 = bs.getModel("rampageBG2")
        cls.railingCollideModel = bs.getCollideModel("rampageBumper")
    
    def __init__(self,activity,retain):
        Map.__init__(self,activity,retain)
        self.node = bs.newNode("terrain",
                               owner=activity,
                               delegate=self,
                               attrs={"collideModel":self.collideModel,
                                      "model":self.model,
                                      "colorTexture":self.tex,
                                      "materials":[bs.footingMaterial]})
        self.bg = bs.newNode("terrain",
                             owner=self.node,
                             attrs={"model":self.bgModel,
                                    "lighting":False,
                                    "background":True,
                                    "colorTexture":self.bgTex})
        self.bottom = bs.newNode("terrain",
                                 owner=self.node,
                                 attrs={"model":self.bottomModel,
                                        "lighting":False,
                                        "colorTexture":self.tex})
        self.bg2 = bs.newNode("terrain",
                              owner=self.node,
                              attrs={"model":self.bgModel2,
                                     "lighting":False,
                                     "background":True,
                                     "colorTexture":self.bgTex2})
        self.railing = bs.newNode("terrain",
                                  owner=self.node,
                                  attrs={"collideModel":self.railingCollideModel,
                                         "materials":[bs.railingMaterial],
                                         "bumper":True})
        bsGlobals = bs.getGlobals()
        bsGlobals.tint = (1.2,1.1,0.97)
        bsGlobals.ambientColor = (1.3,1.2,1.03)
        bsGlobals.vignetteOuter = (0.62,0.64,0.69)
        bsGlobals.vignetteInner = (0.97,0.95,0.93)

    def _isPointNearEdge(self,p,running=False):
        boxTranslate = self.defs.boxes['edgeBox'][0:3]
        boxScale = self.defs.boxes['edgeBox'][6:9]
        x = (p.x() - boxTranslate[0])/boxScale[0]
        z = (p.z() - boxTranslate[2])/boxScale[2]
        return (x < -0.5 or x > 0.5 or z < -0.5 or z > 0.5)

registerMap(RampageMap)
